from nose.tools import with_setup
import jug.task
from jug.backends.dict_store import dict_store

def _setup():
    jug.task.Task.store = dict_store()
def _teardown():
    jug.task.alltasks = []
task_reset = with_setup(_setup, _teardown)



from nose.tools import with_setup
import jug.jug
import jug.task
from jug.task import Task
from jug.backends.dict_store import dict_store
from tests.task_reset import task_reset
import random
jug.jug.silent = True


@task_reset
def test_jug_invalidate():
    def setAi(i):
        A[i] = True
    N = 1024
    A = [False for i in xrange(N)]
    setall = [Task(setAi, i) for i in xrange(N)]
    store = dict_store()
    jug.task.Task.store = store
    for t in setall: t.run()
    jug.jug.invalidate(store, setall[0].name)
    assert not store.store.keys(), store.store.keys()
    jug.task.Task.store = dict_store()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2010, Luis Pedro Coelho <lpc@cmu.edu>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.


from ..task import value

def load_all(jugfile):
    '''
    load_all(jugfile)

    Loads the result of all tasks.
    '''
    for elem in dir(jugfile):
        try:
            v = value(getattr(jugfile, elem))
            setattr(jugfile, elem, v)
        except Exception as e:
            print 'Error while loading %s: %s' % (elem, e)


def shell(store, jugmodule):
    '''
    shell(store, jugmodule)

    Implement 'shell' command.

    Currently depends on Ipython being installed.
    '''
    jugfilename = jugmodule.__name__
    try:
        from IPython.Shell import IPShellEmbed
    except ImportError:
        print >>sys.stderr, '''\
jug: Error: could not import IPython libraries

IPython is necessary for `shell` command.
'''
        sys.exit(1)
    if jugfilename == 'jugfile':
        msg = 'The jugfile is available as `jugfile`.'
    else:
        msg = 'The jugfile is available as `jugfile` and as `%s`.' % jugfilename

    def _load_all():
        '''
        load_all()

        Loads all task results.
        '''
        load_all(jugmodule)

    ipshell = IPShellEmbed(banner='''
=========
Jug Shell
=========

%s

Enjoy...
''' % msg)

    local_ns = {
        'jugfile' : jugmodule,
        jugfilename : jugmodule,
        'load_all' : _load_all,
    }
    ipshell(local_ns=local_ns)





import jug.jug
import jug.task
from jug.task import Task
from jug.backends.dict_store import dict_store
from tests.task_reset import task_reset
import random
jug.jug.silent = True


def test_jug_check():
    N = 1024
    A = [False for i in xrange(N)]
    def setAi(i):
        A[i] = True
    setall = [Task(setAi, i) for i in xrange(N)]
    store = dict_store()
    jug.task.Task.store = store
    e = None
    try:
        jug.jug.check(store)
    except SystemExit, e:
        pass
    assert e is not None
    assert e.code == 1
    jug.jug.execute(store)

    e = None
    try:
        jug.jug.check(store)
    except SystemExit, e:
        pass
    assert e is not None
    assert e.code == 0




import jug.jug
import jug.task
from jug.task import Task
from jug.backends.dict_store import dict_store
import random
jug.jug.silent = True

def test_jug_execute_simple():
    N = 1024
    random.seed(232)
    A = [False for i in xrange(N)]
    def setAi(i):
        A[i] = True
    setall = [Task(setAi, i) for i in xrange(N)]
    store = dict_store()
    jug.task.Task.store = store
    jug.jug.execute(store)
    assert False not in A
    assert max(store.counts.values()) < 4

def test_jug_execute_deps():
    N = 256
    random.seed(234)
    A = [False for i in xrange(N)]
    def setAi(i, other):
        A[i] = True
    idxs = range(N)
    random.shuffle(idxs)
    prev = None
    for idx in idxs:
        prev = Task(setAi, idx, prev)
    store = dict_store()
    jug.task.Task.store = store
    jug.jug.execute(store)
    assert False not in A
    assert max(store.counts.values()) < 4




# -*- coding: utf-8 -*-
# Copyright (C) 2008-2010, Luis Pedro Coelho <lpc@cmu.edu>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
'''
Task: contains the Task class.

This is the main class for using jug.

There are two main alternatives:

- Use the ``Task`` class directly to build up tasks, such as
  ``Task(function, arg0, ...)``.
- Rely on the ``TaskGenerator`` decorator as a shortcut for this.
'''

from __future__ import division
import hashlib
import os
import cPickle as pickle
import itertools

alltasks = []

class Task(object):
    '''
    Task
    ----

    T = Task(f, dep0, dep1,..., kw_arg0=kw_val0, kw_arg1=kw_val1, ...)

    Defines a task, which is roughly equivalent to

    f(dep0, dep1,..., kw_arg0=kw_val0, kw_arg1=kw_val1, ...)

    '''
    store = None
    def __init__(self, f, *args, **kwargs):
        if f.func_name == '<lambda>':
            raise ValueError('''jug.Task does not work with lambda functions!

Write an email to the authors if you feel you have a strong reason to use them (they are a bit
tricky to support since the general code relies on the function name)''')

        self.name = '%s.%s' % (f.__module__, f.__name__)
        self.f = f
        self.args = args
        self.kwargs = kwargs
        self.finished = False
        self.loaded = False
        self._hash = None
        self._lock = None
        self._can_load = False
        alltasks.append(self)

    def run(self, force=False):
        '''
        task.run(force=False)

        Performs the task.

        Parameters
        ----------
          force : if True, always run the task (even if it ran before)
        '''
        assert self.can_run()
        assert force or not self.finished
        args = [value(dep) for dep in self.args]
        kwargs = dict((key,value(dep)) for key,dep in self.kwargs.iteritems())
        self._result = self.f(*args,**kwargs)
        name = self.hash()
        self.store.dump(self._result, name)
        self.finished = True
        return self._result

    def _get_result(self):
        if not self.finished: self.load()
        return self._result

    result = property(_get_result,doc='Result value')


    def can_run(self):
        '''
        bool = task.can_run()

        Returns true if all the dependencies have their results available.
        '''
        for dep in self.dependencies():
            if not dep.finished and not dep.can_load():
                return False
        return True

    def load(self):
        '''
        t.load()

        Loads the results from the storage backend.
        '''
        assert self.can_load()
        self._result = self.store.load(self.hash())
        self.finished = True

    def unload(self):
        '''
        t.unload()

        Unload results (can be useful for saving memory).
        '''
        self._result = None
        self.finished = False

    def unload_recursive(self):
        '''
        t.unload_recursive()

        Equivalent to

        for tt in recursive_dependencies(t): tt.unload()
        '''
        self.unload()
        for dep in self.dependencies():
            dep.unload_recursive()


    def dependencies(self):
        '''
        for dep in task.dependencies():
            ...

        Iterates over all the first-level dependencies of task `t`

        Parameters
        ----------
          self : Task
        Returns
        -------
          A generator over all of `self`'s dependencies
        See Also
        --------
          `recursive_dependencies`
        '''
        queue = [self.args, self.kwargs.values()]
        while queue:
            deps = queue.pop()
            for dep in deps:
                if type(dep) is Task:
                    yield dep
                elif type(dep) in (list,tuple):
                    queue.append(dep)
                elif type(dep) is dict:
                    queue.append(dep.itervalues())


    def can_load(self, store=None):
        '''
        bool = task.can_load()

        Returns whether result is available.
        '''
        if store is None:
            store = self.store
        if self.finished: return True
        if not self._can_load:
            self._can_load = store.can_load(self.hash())
        return self._can_load

    def hash(self):
        '''
        fname = t.hash()

        Returns the hash for this task.

        The results are cached, so the first call can be much slower than
        subsequent calls.
        '''
        if self._hash is None:
            M = hashlib.md5()
            def update(elems):
                for n,e in elems:
                    M.update(pickle.dumps(n))
                    if type(e) == Task:
                        M.update(e.hash())
                    elif type(e) in (list, tuple):
                        update(enumerate(e))
                    elif type(e) == dict:
                        update(e.iteritems())
                    else:
                        M.update(pickle.dumps(e))
            M.update(self.name)
            update(enumerate(self.args))
            update(self.kwargs.iteritems())
            self._hash = M.hexdigest()
        return self._hash

    def __str__(self):
        '''String representation'''
        return 'Task: %s()' % self.name

    def __repr__(self):
        '''Detailed representation'''
        return 'Task(%s, args=%s, kwargs=%s)' % (self.name, self.args, self.kwargs)

    def lock(self):
        '''
        locked = task.lock()

        Tries to lock the task for the current process.

        Returns True if the lock was acquired. The correct usage pattern is::

            locked = task.lock()
            if locked:
                task.run()
            else:
                # someone else is already running this task!

        Not that using can_lock() can lead to race conditions. The above
        is the only fully correct method.

        Returns
        -------
          Whether the lock was obtained.
        '''
        if self._lock is None:
            self._lock = self.store.getlock(self.hash())
        return self._lock.get()

    def unlock(self):
        '''
        task.unlock()

        Releases the lock.

        If the lock was not held, this may remove another thread's lock!
        '''
        self._lock.release()

    def is_locked(self):
        '''
        is_locked = t.is_locked()

        Note that only calling lock() and checking the result atomically checks
        for the lock(). This function can be much faster, though, and, therefore
        is sometimes useful.

        Returns
        -------
            Whether the task **appears** to be locked.

        See Also
        --------
        - lock()
        - unlock()
        '''
        if self._lock is None:
            self._lock = self.store.getlock(self.hash())
        return self._lock.is_locked()


def topological_sort(tasks):
    '''
    topological_sort(tasks)

    Sorts a list of tasks topologically in-place. The list is sorted when
    there is never a dependency between tasks[i] and tasks[j] if i < j.
    '''
    sorted = []
    whites = set(tasks)
    def dfs(t):
        for dep in t.dependencies():
            if dep in whites:
                whites.remove(dep)
                dfs(dep)
        sorted.append(t)
    while whites:
        next = whites.pop()
        dfs(next)
    tasks[:] = sorted

def recursive_dependencies(t, max_level=-1):
    '''
    for dep in recursive_dependencies(t, max_level=-1):
        ...

    Returns a generator that lists all recursive dependencies of task

    Parameters
    ----------
      t : task
      max_level : Maximum recursion depth. Set to -1 or None for no recursion limit.
    Returns
    -------
      A generator
    '''
    if max_level is None:
        max_level = -1
    if max_level == 0:
        return

    for dep in t.dependencies():
        yield dep
        for d2 in recursive_dependencies(dep, max_level - 1):
            yield d2

def value(elem):
    '''
    value = value(task)

    Loads a task object recursively. This correcly handles lists,
    dictonaries and eny other type handled by the tasks themselves.
    '''
    if type(elem) is Task:
        return elem.result
    elif type(elem) is list:
        return [value(e) for e in elem]
    elif type(elem) is tuple:
        return tuple([value(e) for e in elem])
    elif type(elem) is dict:
        return dict((x,value(y)) for x,y in elem.iteritems())
    else:
        return elem

def CachedFunction(f,*args,**kwargs):
    '''
    value = CachedFunction(f, *args, **kwargs)

    is equivalent to:

    task = Task(f, *args, **kwargs)
    if not task.can_load():
        task.run()
    value = task.result

    That is, it calls the function if the value is available,
    but caches the result for the future.
    '''
    t = Task(f,*args, **kwargs)
    if not t.can_load():
        if not t.can_run():
            raise ValueError('jug.CachedFunction: unable to run task %s' % t)
        t.run()
    return t.result

def TaskGenerator(f):
    '''
    @TaskGenerator
    def f(arg0, arg1, ...)
        ...

    Turns f from a function into a task generator.

    This means that calling ``f(arg0, arg1)`` results in:
    ``Task(f, arg0, arg1)``
    '''
    from functools import wraps
    @wraps(f)
    def task_generator(*args, **kwargs):
        return Task(f, *args, **kwargs)
    task_generator.__name__ = ('TaskGenerator(%s)' % task_generator.__name__)
    return task_generator 




import jug.backends.redis_store
import jug.backends.file_store
import jug.backends.dict_store
import pickle

_storedir = 'jugtests'
def _remove_file_store():
    jug.backends.file_store.file_store.remove_store(_storedir)

def test_stores():
    def test_load_get(store):
        key = 'jugisbestthingever'
        assert not store.can_load(key)
        object = range(232)
        store.dump(object, key)
        assert store.can_load(key)
        assert store.load(key) == object
        store.remove(key)
        assert not store.can_load(key)
        store.close()

    def test_lock(store):
        key = 'jugisbestthingever'
        lock = store.getlock(key)
        assert not lock.is_locked()
        assert lock.get()
        assert not lock.get()
        lock2 = store.getlock(key)
        assert not lock2.get()
        lock.release()
        assert lock2.get()
        lock2.release()
        store.close()

    functions = (test_load_get, test_lock)
    stores = (lambda: jug.redis_store.redis_store('redis:'), lambda: jug.backends.file_store.file_store('jugtests'), jug.backends.dict_store.dict_store)
    teardowns = (None, _remove_file_store, None)
    for f in functions:
        for s,tear in zip(stores,teardowns):
            f.teardown = tear
            yield f, s()




# -*- coding: utf-8 -*-
# Copyright (C) 2008-2010, Luis Pedro Coelho <lpc@cmu.edu>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
'''
Options

Variables
---------
- jugdir: main jug directory.
- jugfile: filesystem name for the Jugfile
- cmd: command to run.
- aggressive_unload: --aggressive-unload
- invalid_name: --invalid
- argv: Arguments not captured by jug (for script use)
- print_out: Print function to be used for output (behaves like Python3's print)
'''
from __future__ import division
import logging
import string
import sys

from .p3 import nprint

jugdir = 'jugdata'
jugfile = 'jugfile.py'
cmd = None
aggressive_unload = False
invalid_name = None
argv = None
print_out = nprint
status_mode = 'no-cached'

_Commands = (
    'execute',
    'status',
    'check',
    'stats',
    'cleanup',
    'count',
    'invalidate',
    'shell',
    )
_Usage_string = \
'''python %s COMMAND JUGFILE OPTIONS...

Commands:
   execute:      Execute tasks
   status:       Print status
   check:        Returns 0 if all tasks are finished. 1 otherwise.
   counts:       Simply count tasks
   cleanup:      Cleanup
   invalidate:   Invalidate the results of a task
   shell:        Run a shell after initialization

Options:
--aggressive-unload
    Aggressively unload data from memory
--invalid=TASK-NAME
    Task name to invalidate (for invalidate command only)
--jugdir=JUGDIR
    Directory in which to save intermediate files
--verbose=LEVEL
    Verbosity level ('DEBUG', 'INFO', 'QUIET')
'''

def usage():
    '''
    usage()

    Print an usage string and exit.
    '''
    import sys
    print _Usage_string % sys.argv[0]
    sys.exit(1)

def parse():
    '''
    options.parse()

    Parse the command line options and set the option variables.
    '''
    import optparse
    global jugdir, jugfile, cmd, aggressive_unload, invalid_name, argv, status_mode
    parser = optparse.OptionParser()
    parser.add_option('--aggressive-unload',action='store_true',dest='aggressive_unload',default=False)
    parser.add_option('--invalid',action='store',dest='invalid_name',default=None)
    parser.add_option('--jugdir',action='store',dest='jugdir',default='jugdata/')
    parser.add_option('--verbose',action='store',dest='verbosity',default='QUIET')
    parser.add_option('--cache', action='store_true', dest='cache', default=False)
    options,args = parser.parse_args()
    if not args:
        usage()
        return

    cmd = args.pop(0)
    jugfile = 'jugfile.py'
    if args:
        jugfile = args.pop(0)

    if cmd not in _Commands:
        usage()
        return
    if options.invalid_name and cmd != 'invalidate':
        usage()
        return
    if cmd == 'invalidate' and not options.invalid_name:
        usage()
        return
    try:
        nlevel = {
            'DEBUG' : logging.DEBUG,
            'INFO' : logging.INFO,
        }[string.upper(options.verbosity)]
        root = logging.getLogger()
        root.level = nlevel
    except KeyError:
        pass

    aggressive_unload = options.aggressive_unload
    invalid_name = options.invalid_name
    argv = args
    sys.argv = [jugfile] + args
    status_mode = ('cached' if options.cache else 'no-cached')
    jugdir = options.jugdir


# vim: set ts=4 sts=4 sw=4 expandtab smartindent:



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2010, Luis Pedro Coelho <lpc@cmu.edu>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

from __future__ import division
from collections import defaultdict
from signal import signal, SIGTERM
from time import sleep
import sys
import os
import os.path
import random
import logging

from . import options
from . import task
from . import backends
from .backends import memoize_store
from .task import Task
from .subcommands.status import status
from .subcommands.shell import shell
from .options import print_out

def do_print(store):
    '''
    do_print(store)

    Print a count of task names.
    '''
    task_counts = defaultdict(int)
    for t in task.alltasks:
        task_counts[t.name] += 1
    for tnc in task_counts.items():
        print_out('Task %s: %s' % tnc)

def invalidate(store, invalid_name):
    '''
    invalidate()

    Implement 'invalidate' command
    '''
    invalid = set()
    tasks = task.alltasks
    cache = {}
    def isinvalid(t):
        h = t.hash()
        if h in cache:
            return cache[h]
        if t.name == invalid_name:
            cache[h] = True
            return True
        for dep in t.dependencies():
            if isinvalid(dep):
                cache[h] = True
                return True
        cache[h] = False
        return False

    invalid = filter(isinvalid, tasks)
    if not invalid:
        print_out('No results invalidated.')
        return
    task_counts = defaultdict(int)
    for t in invalid:
        if store.remove(t.hash()):
            task_counts[t.name] += 1
    if sum(task_counts.values()) == 0:
        print_out('Tasks invalidated, but no results removed')
    else:
        print_out('Tasks Invalidated')
        print_out()
        print_out('Task Name                                   Count')
        print_out('-------------------------------------------------')
        #          0         1         2         3         4         5
        #          012345678901234567890123456789012345678901234567890123456789
        for n_c in task_counts.items():
            print_out('%-40s: %7s' % n_c)


def execute(store, aggressive_unload=False):
    '''
    execute(store, aggressive_unload=False)

    Implement 'execute' command
    '''

    tasks = task.alltasks
    task_names = set(t.name for t in tasks)
    tasks_executed = defaultdict(int)
    tasks_loaded = defaultdict(int)
    task.topological_sort(tasks)
    logging.info('Execute start (%s tasks)' % len(tasks))
    signal(SIGTERM,_sigterm)
    while tasks:
        upnext = []
        for i in xrange(30*60//12): #This is at most half-an-hour
            cannot_run = 0
            max_cannot_run = min(len(tasks), 128)
            while not tasks[0].can_run() and cannot_run < max_cannot_run:
                tasks.append(tasks.pop(0))
                cannot_run += 1
            while tasks and tasks[0].can_run():
                upnext.append(tasks.pop(0))
            if upnext:
                break
            logging.info('waiting 12 secs for an open task...')
            sleep(12)
        if not upnext:
            logging.info('No tasks can be run!')
            return
        for t in upnext:
            if t.can_load():
                logging.info('Loadable %s...' % t.name)
                tasks_loaded[t.name] += 1
                continue
            locked = False
            try:
                locked = t.lock()
                if t.can_load(): # This can be true if the task ran between the check above and this one
                    logging.info('Loadable %s...' % t.name)
                    tasks_loaded[t.name] += 1
                elif locked:
                    logging.info('Executing %s...' % t.name)
                    t.run()
                    tasks_executed[t.name] += 1
                    if aggressive_unload:
                        t.unload_recursive()
                else:
                    logging.info('Already in execution %s...' % t.name)
            except Exception, e:
                logging.critical('Exception while running %s: %s' % (t.name,e))
                raise
            finally:
                if locked: t.unlock()

    print_out('%-52s%12s%12s' %('Task name','Executed','Loaded'))
    print_out('-' * (52+12+12))
    for t in task_names:
        name_cut = t[:52]
        print_out('%-52s%12s%12s' % (name_cut,tasks_executed[t],tasks_loaded[t]))
    if not task_names:
        print_out('<no tasks>')


def cleanup(store):
    '''
    cleanup(store)

    Implement 'cleanup' command
    '''
    tasks = task.alltasks
    removed = store.cleanup(tasks)
    print_out('Removed %s files' % removed)


def check(store):
    '''
    check(store)

    Executes check subcommand
    '''
    tasks = task.alltasks
    while tasks:
        t = tasks.pop()
        if not t.can_load(store):
            sys.exit(1)
        else:
            for dep in t.dependencies():
                if dep in tasks:
                    del tasks[tasks.index(dep)]
    sys.exit(0)


def init(jugfile=None, jugdir=None, on_error='exit'):
    '''
    store = init(jugfile={'jugfile'}, jugdir={'jugdata'}, on_error='exit')

    Initializes jug (create backend connection, ...).
    Imports jugfile

    Parameters
    ----------
    `jugfile` : jugfile to import (default: 'jugfile')
    `jugdir` : jugdir to use (could be a path)
    `on_error` : What to do if import fails (default: exit)

    Returns
    -------
    `store` : storage object
    '''
    assert on_error in ('exit', 'propagate'), 'jug.init: on_error option is not valid.'
    if jugfile is None:
        jugfile = 'jugfile'
    if jugdir is None:
        jugdir = 'jugdata'
    store = backends.select(jugdir)
    Task.store = store

    if jugfile.endswith('.py'):
        jugfile = jugfile[:-len('.py')]
    sys.path.insert(0, os.path.abspath('.'))
    try:
        jugmodule = __import__(jugfile)
    except ImportError, e:
        logging.critical("Could not import file '%s' (error: %s)", jugfile, e)
        if on_error == 'exit':
            sys.exit(1)
        else:
            raise
    return store, jugmodule


def _sigterm(_,__):
    sys.exit(1)

def main():
    options.parse()
    if options.cmd != 'status':
        store,jugmodule = init(options.jugfile, options.jugdir)

    if options.cmd == 'execute':
        execute(store, options.aggressive_unload)
    elif options.cmd == 'count':
        do_print(store)
    elif options.cmd == 'check':
        check(store)
    elif options.cmd == 'status':
        status()
    elif options.cmd == 'invalidate':
        invalidate(store, options.invalid_name)
    elif options.cmd == 'cleanup':
        cleanup(store)
    elif options.cmd == 'shell':
        shell(store, jugmodule)
    else:
        logging.critical('Jug: unknown command: \'%s\'' % options.cmd)

if __name__ == '__main__':
    try:
        main()
    except Exception, exc:
        logging.critical('Unhandled Jug Error!')
        raise



from jug.backends.encode import encode, decode
import numpy as np
def test_encode():
    assert decode(encode(None)) is None
    assert decode(encode([])) == []
    assert decode(encode(range(33))) == range(33)
    assert np.all(decode(encode(np.arange(33))) == np.arange(33))




import jug.mapreduce
import numpy as np
from jug.backends.dict_store import dict_store

def mapper(x):
    return x**2
def reducer(x, y):
    return x + y
def dfs_run(t):
    for dep in t.dependencies():
        dfs_run(dep)
    t.run()

def test_mapreduce():
    np.random.seed(33)
    jug.task.Task.store = dict_store()
    A = np.random.rand(10000)
    t = jug.mapreduce.mapreduce(reducer, mapper, A)
    dfs_run(t)
    assert np.abs(t.result - (A**2).sum()) < 1.




#-*- coding: utf-8 -*-
# Copyright (C) 2009-2010, Luis Pedro Coelho <lpc@cmu.edu>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

'''
redis_store: store based on a redis backend
'''
from __future__ import division

import re
import logging
from jug.backends.encode import encode, decode
from base64 import b64encode, b64decode

try:
    import redis
except ImportError:
    try:
        from ..thirdparty import redis
    except ImportError:
        redis = None

def _resultname(name):
    return 'result:' + name

def _lockname(name):
    return 'lock:' + name

_LOCKED = 1

_redis_urlpat = re.compile('redis://(?P<host>[A-Za-z0-9\.]+)(?P<port>\:[0-9]+)?/')


class redis_store(object):
    def __init__(self, url):
        '''
        '''
        if redis is None:
            raise IOError, 'jug.redis_store: redis module is not found!'
        redis_params = {}
        match = _redis_urlpat.match(url)
        if match:
            redis_params = match.groupdict()
        logging.info('connecting to %s' % redis_params)

        self.redis = redis.Redis(**redis_params)


    def dump(self, object, name):
        '''
        dump(object, name)
        '''
        s = encode(object)
        if s:
            s = b64encode(s)
        self.redis.set(_resultname(name), s)


    def can_load(self, name):
        '''
        can = can_load(name)
        '''
        return self.redis.exists(_resultname(name))


    def load(self, name):
        '''
        obj = load(name)

        Loads the object identified by `name`.
        '''
        s = self.redis.get(_resultname(name))
        if s:
            s = str(s)
            s = b64decode(s)
        return decode(s)


    def remove(self, name):
        '''
        was_removed = remove(name)

        Remove the entry associated with name.

        Returns whether any entry was actually removed.
        '''
        return self.redis.delete(_resultname(name))


    def cleanup(self, active):
        '''
        cleanup()

        Implement 'cleanup' command
        '''
        existing = self.redis.keys('result:*')
        for act in active:
            try:
                existing.remove(_resultname(act))
            except KeyError:
                pass
        for superflous in existing:
            self.redis.delete(_resultname(superflous))


    def getlock(self, name):
        return redis_lock(self.redis, name)


    def close(self):
        self.redis.disconnect()



class redis_lock(object):
    '''
    redis_lock

    Functions:
    ----------

        * get(): acquire the lock
        * release(): release the lock
        * is_locked(): check lock state
    '''

    def __init__(self, redis, name):
        self.name = _lockname(name)
        self.redis = redis


    def get(self):
        '''
        lock.get()
        '''
        previous = self.redis.getset(self.name, _LOCKED)
        return (previous is None)


    def release(self):
        '''
        lock.release()

        Removes lock
        '''
        self.redis.delete(self.name)


    def is_locked(self):
        '''
        locked = lock.is_locked()
        '''
        status = self.redis.get(self.name)
        return status is not None and status == _LOCKED




from jug.backends.file_store import file_store, file_based_lock
from nose.tools import with_setup

_storedir = 'jugtests'
def _remove_file_store():
    file_store.remove_store(_storedir)

@with_setup(teardown=_remove_file_store)
def test_twice():
    lock = file_based_lock(_storedir, 'foo')
    assert lock.get()
    assert not lock.get()
    lock.release()

    assert lock.get()
    assert not lock.get()
    lock.release()

@with_setup(teardown=_remove_file_store)
def test_twolocks():
    foo = file_based_lock(_storedir, 'foo')
    bar = file_based_lock(_storedir, 'bar')
    assert foo.get()
    assert bar.get()
    assert not foo.get()
    assert not bar.get()
    foo.release()
    bar.release()




# -*- coding: utf-8 -*-
# Copyright (C) 2008-2010, Luis Pedro Coelho <lpc@cmu.edu>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

from __future__ import division
from jug import Task
from itertools import chain

def _jug_map(mapper, inputs):
     return map(mapper, inputs)
def _jug_reduce(reducer, inputs):
    return reduce(reducer, chain(inputs))

def mapreduce(reducer, mapper, inputs, map_step=4, reduce_step=8):
    '''
    task = mapreduce(reducer, mapper, inputs, map_step=4, reduce_step=8)

    Create a task that does roughly the following::

        reduce(reducer, map(mapper, inputs))

    Roughly because the order of operations might be different. In particular,
    `reducer` should be a true `reducer` functions (i.e., cumutative and
    associative).

    Parameters
    ----------
      reducer : associative, comutative function mapping
                    [Y_0,Y_1...] -> Y
      mapper : function from X -> Y
      inputs : list of X
    
      map_step : Number of mapping operations to do in one go.
                 This is what defines an inner task. (default: 4)
      reduce_step : Number of reduce operations to do in one go.
                 (default: 8)
    Returns
    -------
      A task
    '''
    def _task_map(jug_op, func, args, step):
        nsteps = len(args)//step + bool(len(args) % step)
        return [Task(jug_op, func, args[i*step:(i+1)*step]) for i in xrange(nsteps)]
    reducers = [Task(_jug_reduce, reducer, mapped) for mapped in _task_map(_jug_map, mapper, inputs, map_step)]
    while len(reducers) > 1:
        reducers = _task_map(_jug_reduce, reducer, reducers, reduce_step)
    return reducers[0]




#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2010, Luis Pedro Coelho <lpc@cmu.edu>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

from collections import defaultdict
import sqlite3

import jug
from ..task import recursive_dependencies
from .. import task
from .. import backends
from .. import options
from ..task import Task
from ..options import print_out
from ..backends import memoize_store

unknown,waiting,ready,running,finished = range(5)
_jugstatus_file = '.jugstatus.sqlite3'

def _create_sqlite3(ht, deps, rdeps):
    connection = sqlite3.connect(_jugstatus_file)
    connection.execute('''
    CREATE TABLE ht (
            id INTEGER PRIMARY KEY,
            name CHAR(128),
            hash CHAR(128),
            status INT );
    ''')
    connection.execute('''
    CREATE TABLE dep (
        source INT,
        target INT);
    ''')

    for h in ht:
        connection.execute('''
       INSERT INTO ht(name, hash, status) VALUES(?,?,?)
    ''', (h[1], h[2], h[3]))

    for i,cdeps in deps.iteritems():
        for cd in cdeps:
            connection.execute('''
           INSERT INTO dep(source, target) VALUES(?,?)
            ''', (i,cd))
    connection.commit()
    connection.close()

def _retrieve_sqlite3():
    connection = sqlite3.connect(_jugstatus_file)
    cursor = connection.execute('''SELECT * FROM ht''')
    ht = cursor.fetchall()
    cursor = connection.execute('''SELECT * FROM dep''')
    deps = defaultdict(list)
    rdeps = defaultdict(list)
    for d0,d1 in cursor:
        deps[d0].append(d1)
        rdeps[d1].append(d0)
    return ht, deps, rdeps

def _save_dirty3(dirty):
    connection = sqlite3.connect(_jugstatus_file)
    for id,status in dirty:
        connection.execute('''UPDATE ht SET STATUS = ? WHERE id = ?''', (status, id))
    connection.commit()


def _load_jugfile():
    store,_ = jug.init(options.jugfile, options.jugdir)
    h2idx = {}
    ht = []
    deps = {}
    rdeps = {}
    for i,t in enumerate(task.alltasks):
        hash = t.hash()
        curdeps = [h2idx[dep.hash()] for dep in recursive_dependencies(t, 1)]
        if curdeps:
            deps[i] = curdeps
        h = [i, t.name, hash, unknown]
        ht.append(h)
        h2idx[hash] = i

    for k,v in deps.iteritems():
        for rv in v:
            if rv not in rdeps:
                rdeps[rv] = [k]
            else:
                rdeps[rv].append(k)
    return store, ht, deps, rdeps


def _update_status(store, ht, deps, rdeps):
    tasks_waiting = defaultdict(int)
    tasks_ready = defaultdict(int)
    tasks_running = defaultdict(int)
    tasks_finished = defaultdict(int)

    store = memoize_store(store)
    dirty = []
    active = []
    for h in ht:
        _,name,hash,status = h
        if status == finished:
            tasks_finished[name] += 1
        else:
            active.append(h)

    for i,name,hash,status in active:
        nstatus = None
        if store.can_load(hash):
            tasks_finished[name] += 1
            nstatus = finished
        else:
            can_run = True
            if status != ready:
                for dep in deps.get(i, []):
                    _,_,dhash,dstatus = ht[dep]
                    if dstatus != finished and not store.can_load(dhash):
                        can_run = False
                        break
            if can_run:
                lock = store.getlock(hash)
                if lock.is_locked():
                    tasks_running[name] += 1
                    nstatus = running
                else:
                    tasks_ready[name] += 1
                    nstatus = ready
            else:
                tasks_waiting[name] += 1
                nstatus = waiting
        assert nstatus is not None, '_update_status: nstatus not assigned'
        if status != nstatus:
            dirty.append((i,nstatus))
    return tasks_waiting, tasks_ready, tasks_running, tasks_finished, dirty


def _print_status(waiting, ready, running, finished):
    names = set(waiting.keys())
    names.update(ready.keys())
    names.update(running.keys())
    names.update(finished.keys())
    format      = '%-40s%12s%12s%12s%12s'
    format_size =    40 +12 +12 +12 +12
    print_out(format % ('Task name','Waiting','Ready','Finished','Running'))
    print_out('-' * format_size)

    for n in names:
        n_cut = n[:40]
        print_out(format % (n_cut,waiting[n],ready[n],finished[n],running[n]))
    print_out('.' * format_size)
    print_out(format % ('Total:',sum(waiting.values()),sum(ready.values()),sum(finished.values()),sum(running.values())))
    print_out()


def _status_cached():
    create, update = range(2)
    try:
        ht, deps, rdeps = _retrieve_sqlite3()
        store = backends.select(options.jugdir)
        mode = update
    except:
        store, ht, deps, rdeps = _load_jugfile()
        mode = create

    tw,tre,tru,tf,dirty = _update_status(store, ht, deps, rdeps)
    _print_status(tw, tre, tru, tf)
    if mode == update:
        _save_dirty3(dirty)
    else:
        for i,nstatus in dirty:
            _,name,hash,_ = ht[i]
            ht[i] = (i, name, hash, nstatus)
        _create_sqlite3(ht, deps, rdeps)


def _status_nocache():
    store,jugmodule = jug.init(options.jugfile, options.jugdir)
    Task.store = memoize_store(store)

    task_names = set(t.name for t in task.alltasks)
    tasks_waiting = defaultdict(int)
    tasks_ready = defaultdict(int)
    tasks_running = defaultdict(int)
    tasks_finished = defaultdict(int)
    for t in task.alltasks:
        if t.can_load():
            tasks_finished[t.name] += 1
        elif t.can_run():
            if t.is_locked():
                tasks_running[t.name] += 1
            else:
                tasks_ready[t.name] += 1
        else:
            tasks_waiting[t.name] += 1
    _print_status(tasks_waiting, tasks_ready, tasks_running, tasks_finished)


def status():
    '''
    status(store)

    Implements the status command.
    '''
    if options.status_mode == 'cached':
        _status_cached()
    else:
        _status_nocache()
